@@ -2,7 +2,7 @@
 PyTorch Forecasting package for timeseries forecasting with PyTorch.
 """
 from pytorch_forecasting.data import EncoderNormalizer, GroupNormalizer, TimeSeriesDataSet
-from pytorch_forecasting.models import Baseline, NBeats, TemporalFusionTransformer
+from pytorch_forecasting.models import Baseline, DeepAR, NBeats, TemporalFusionTransformer
 
 __all__ = [
     "TimeSeriesDataSet",
@@ -11,6 +11,7 @@ __all__ = [
     "TemporalFusionTransformer",
     "NBeats",
     "Baseline",
+    "DeepAR",
 ]
 
-__version__ = "0.5.1"
+__version__ = "0.7.0"
