@@ -115,26 +115,48 @@ class TimeSeriesDataSet(Dataset):
         add_relative_time_idx: bool = False,
         add_target_scales: bool = False,
         add_encoder_length: Union[bool, str] = "auto",
-        target_normalizer: Union[TorchNormalizer, NaNLabelEncoder, str] = "auto",
+        target_normalizer: Union[TorchNormalizer, NaNLabelEncoder, EncoderNormalizer, str] = "auto",
         categorical_encoders={},
         scalers={},
         randomize_length: Union[None, Tuple[float, float], bool] = False,
         predict_mode: bool = False,
     ):
         """
-        Timeseries dataset
+        Timeseries dataset holding data for models.
+
+        Each sample is a subsequence of a full time series. The subsequence consists of encoder and decoder/prediction
+        timepoints for a given time series. This class constructs an index which defined which subsequences exists and
+        can be samples from (``index`` attribute). The samples in the index are defined by by the various parameters.
+        to the class (encoder and prediction lengths, minimum prediction length, randomize length and predict keywords).
+        How samples are
+        sampled into batches for training, is determined by the DataLoader. The class provides the
+        :py:meth:`~TimeSeriesDataSet.to_dataloader` method to convert the dataset into a dataloader.
+
+        Large datasets:
+
+            Currently the class is limited to in-memory operations. If you have extremely large data,
+            however, you can pass prefitted encoders and and scalers to it and a subset of sequences to the class to
+            construct a valid dataset (plus, likely the EncoderNormalizer should be used to normalize targets).
+            when fitting a network, you would then to create a custom DataLoader that rotates through the datasets.
+            There is currently no in-built methods to do this.
 
         Args:
             data: dataframe with sequence data - each row can be identified with ``time_idx`` and the ``group_ids``
-            time_idx: integer column denoting the time index
+            time_idx: integer column denoting the time index. This columns is used to determine the sequence of samples.
+                If there no missings observations, the time index should increase by ``+1`` for each subsequent sample.
+                The first time_idx for each series does not necessarily have to be ``0`` but any value is allowed.
             target: column denoting the target or list of columns denoting the target - categorical or continous.
-            group_ids: list of column names identifying a timeseries
+            group_ids: list of column names identifying a time series. This means that the ``group_ids`` identify
+                a sample together with the ``time_idx``. If you have only one timeseries, set this to the
+                name of column that is constant.
             weight: column name for weights or list of column names corresponding to each target
             max_encoder_length: maximum length to encode
             min_encoder_length: minimum allowed length to encode. Defaults to max_encoder_length.
-            min_prediction_idx: minimum time index from where to start predictions
-            max_prediction_length: maximum prediction length (choose this not too short as it can help convergence)
-            min_prediction_length: minimum prediction length. Defaults to max_prediction_length
+            min_prediction_idx: minimum ``time_idx`` from where to start predictions. This parameter can be useful to
+                create a validation or test set.
+            max_prediction_length: maximum prediction/decoder length (choose this not too short as it can help
+                convergence)
+            min_prediction_length: minimum prediction/decoder length. Defaults to max_prediction_length
             static_categoricals: list of categorical variables that do not change over time,
                 entries can be also lists which are then encoded together
                 (e.g. useful for product categories)
@@ -154,23 +176,44 @@ class TimeSeriesDataSet(Dataset):
             dropout_categoricals: list of categorical variables that are unknown when making a forecast without
                 observed history
             constant_fill_strategy: dictionary of column names with constants to fill in missing values if there are
-                gaps in the sequence
-                (otherwise forward fill strategy is used)
-            allow_missings: if to allow missing timesteps that are automatically filled up
-            add_relative_time_idx: if to add a relative time index as feature
-            add_target_scales: if to add scales for target to static real features
+                gaps in the sequence (by default forward fill strategy is used). The values will be only used if
+                ``allow_missings=True``. A common use case is to denote that demand was 0 if the sample is not in
+                the dataset.
+            allow_missings: if to allow missing timesteps that are automatically filled up. Missing values
+                refer to gaps in the ``time_idx``, e.g. if a specific timeseries has only samples for
+                1, 2, 4, 5, the sample for 3 will be generated on-the-fly.
+                Allow missings does not deal with ``NA`` values. You should fill NA values before
+                passing the dataframe to the TimeSeriesDataSet.
+            add_relative_time_idx: if to add a relative time index as feature (i.e. for each sampled sequence, the index
+                will range from -encoder_length to prediction_length)
+            add_target_scales: if to add scales for target to static real features (i.e. add the center and scale
+                of the unnormalized timeseries as features)
             add_encoder_length: if to add decoder length to list of static real variables. Defaults to "auto",
                 i.e. yes if ``min_encoder_length != max_encoder_length``.
-            target_normalizer: transformer that takes group_ids, target and time_idx to return normalized target
-            categorical_encoders: dictionary of scikit learn label transformers or None
-            scalers: dictionary of scikit learn scalers or None
+            target_normalizer: transformer that takes group_ids, target and time_idx to return normalized targets.
+                You can choose from :py:class:`~TorchNormalizer`, :py:class:`~NaNLabelEncoder`,
+                :py:class:`~EncoderNormalizer` or `None` for using not normalizer.
+                By default an appropriate normalizer is chosen automatically.
+            categorical_encoders: dictionary of scikit learn label transformers. If you have unobserved categories in
+                the future, you can use the :py:class:`~pytorch_forecasting.encoders.NaNLabelEncoder` with
+                ``add_nan=True``. Defaults effectively to sklearn's ``LabelEncoder()``. Prefittet encoders will not
+                be fit again.
+            scalers: dictionary of scikit learn scalers. Defaults to sklearn's ``StandardScaler()``.
+                Other options are :py:class:`~pytorch_forecasting.data.EncoderNormalizer`,
+                :py:class:`~pytorch_forecasting.data.GroupNormalizer` or scikit-learn's ``StandarScaler()``,
+                ``RobustScaler()`` or `None` for using not normalizer.
+                Prefittet encoders will not be fit again (with the exception of the
+                :py:class:`~pytorch_forecasting.data.EncoderNormalizer`).
             randomize_length: None or False if not to randomize lengths. Tuple of beta distribution concentrations
                 from which
                 probabilities are sampled that are used to sample new sequence lengths with a binomial
                 distribution.
                 If True, defaults to (0.2, 0.05), i.e. ~1/4 of samples around minimum encoder length.
                 Defaults to False otherwise.
-            predict_mode: if to only iterate over each timeseries once (only the last provided samples)
+            predict_mode: if to only iterate over each timeseries once (only the last provided samples).
+                Effectively, this will take choose for each time series identified by ``group_ids``
+                the last ``max_prediction_length`` samples of each time series as
+                prediction samples and everthing previous up to ``max_encoder_length`` samples as encoder samples.
         """
         super().__init__()
         self.max_encoder_length = max_encoder_length
@@ -298,16 +341,20 @@ class TimeSeriesDataSet(Dataset):
                 self.target_normalizer = NaNLabelEncoder()
                 assert not self.add_target_scales, "Target scales can be only added for continous targets"
             else:
-                coerce_positive = (data[self.target] >= 0).all()
-                if coerce_positive:
-                    log_scale = data[self.target].skew() > 2.5
-                    coerce_positive = False
+                data_positive = (data[self.target] > 0).all()
+                if data_positive:
+                    if data[self.target].skew() > 2.5:
+                        transformer = "log"
+                    else:
+                        transformer = "relu"
                 else:
-                    log_scale = False
+                    transformer = None
                 if self.max_encoder_length > 20 and self.min_encoder_length > 1:
-                    self.target_normalizer = EncoderNormalizer(coerce_positive=coerce_positive, log_scale=log_scale)
+                    self.target_normalizer = EncoderNormalizer(transformation=transformer)
                 else:
-                    self.target_normalizer = GroupNormalizer(coerce_positive=coerce_positive, log_scale=log_scale)
+                    self.target_normalizer = GroupNormalizer(transformation=transformer)
+        elif self.target_normalizer is None:
+            self.target_normalizer = TorchNormalizer(method="identity")
         assert self.min_encoder_length > 1 or not isinstance(
             self.target_normalizer, EncoderNormalizer
         ), "EncoderNormalizer is only allowed if min_encoder_length > 1"
@@ -315,6 +362,24 @@ class TimeSeriesDataSet(Dataset):
             self.target_normalizer, (TorchNormalizer, NaNLabelEncoder)
         ), f"target_normalizer has to be either None or of class TorchNormalizer but found {self.target_normalizer}"
 
+    @property
+    def _group_ids_mapping(self) -> Dict[str, str]:
+        """
+        Mapping of group id names to group ids used to identify series in dataset -
+        group ids can also be used for target normalizer.
+        The former can change from training to validation and test dataset while the later must not.
+        """
+        return {name: f"__group_id__{name}" for name in self.group_ids}
+
+    @property
+    def _group_ids(self) -> List[str]:
+        """
+        Group ids used to identify series in dataset.
+
+        See :py:meth:`~TimeSeriesDataSet._group_ids_mapping` for details.
+        """
+        return list(self._group_ids_mapping.values())
+
     def _validate_data(self, data: pd.DataFrame):
         """
         Validate that data will not cause hick-ups later on.
@@ -330,6 +395,12 @@ class TimeSeriesDataSet(Dataset):
                 raise ValueError(
                     f"Data type of category {name} was found to be numeric - use a string type / categorified string"
                 )
+        # check for "." in column names
+        columns_with_dot = data.columns[data.columns.str.contains(r"\.")]
+        if len(columns_with_dot) > 0:
+            raise ValueError(
+                f"column names must not contain '.' characters. Names {columns_with_dot.tolist()} are invalid"
+            )
 
     def save(self, fname: str) -> None:
         """
@@ -365,9 +436,19 @@ class TimeSeriesDataSet(Dataset):
         Returns:
             pd.DataFrame: pre-processed dataframe
         """
+        # encode group ids - this encoding
+        for name, group_name in self._group_ids_mapping.items():
+            self.categorical_encoders[group_name] = NaNLabelEncoder().fit(data[name].to_numpy().reshape(-1))
+            data[group_name] = self.transform_values(name, data[name], inverse=False, group_id=True)
 
         # encode categoricals
-        for name in set(self.categoricals + self.group_ids):
+        if isinstance(
+            self.target_normalizer, GroupNormalizer
+        ):  # if we use a group normalizer, group_ids must be encoded as well
+            group_ids_to_encode = self.group_ids
+        else:
+            group_ids_to_encode = []
+        for name in set(group_ids_to_encode + self.categoricals):
             allow_nans = name in self.dropout_categoricals
             if name in self.variable_groups:  # fit groups
                 columns = self.variable_groups[name]
@@ -392,7 +473,7 @@ class TimeSeriesDataSet(Dataset):
                         self.categorical_encoders[name] = self.categorical_encoders[name].fit(data[name])
 
         # encode them
-        for name in set(self.flat_categoricals + self.group_ids):
+        for name in set(group_ids_to_encode + self.flat_categoricals):
             data[name] = self.transform_values(name, data[name], inverse=False)
 
         # save special variables
@@ -423,8 +504,9 @@ class TimeSeriesDataSet(Dataset):
                 # transformation over the entire time range but by each group
                 common_init_args = [
                     name
-                    for name in inspect.signature(GroupNormalizer).parameters.keys()
-                    if name in inspect.signature(EncoderNormalizer).parameters.keys()
+                    for name in inspect.signature(GroupNormalizer.__init__).parameters.keys()
+                    if name in inspect.signature(EncoderNormalizer.__init__).parameters.keys()
+                    and name not in ["data", "self"]
                 ]
                 copy_kwargs = {name: getattr(self.target_normalizer, name) for name in common_init_args}
                 normalizer = GroupNormalizer(groups=self.group_ids, **copy_kwargs)
@@ -433,6 +515,10 @@ class TimeSeriesDataSet(Dataset):
                 data[self.target], scales = self.target_normalizer.transform(data[self.target], data, return_norm=True)
             elif isinstance(self.target_normalizer, NaNLabelEncoder):
                 data[self.target] = self.target_normalizer.transform(data[self.target])
+                data["__target__"] = data[
+                    self.target
+                ]  # overwrite target because it requires encoding (continuous targets should not be normalized)
+                scales = "no target scales available for categorical target"
             else:
                 data[self.target], scales = self.target_normalizer.transform(data[self.target], return_norm=True)
 
@@ -449,6 +535,8 @@ class TimeSeriesDataSet(Dataset):
 
         if self.target in self.reals:
             self.scalers[self.target] = self.target_normalizer
+        else:
+            self.categorical_encoders[self.target] = self.target_normalizer
 
         # rescale continuous variables apart from target
         for name in self.reals:
@@ -476,7 +564,12 @@ class TimeSeriesDataSet(Dataset):
         return data
 
     def transform_values(
-        self, name: str, values: Union[pd.Series, torch.Tensor, np.ndarray], data: pd.DataFrame = None, inverse=False
+        self,
+        name: str,
+        values: Union[pd.Series, torch.Tensor, np.ndarray],
+        data: pd.DataFrame = None,
+        inverse=False,
+        group_id: bool = False,
     ) -> np.ndarray:
         """
         Scale and encode values.
@@ -487,12 +580,16 @@ class TimeSeriesDataSet(Dataset):
             data (pd.DataFrame, optional): extra data used for scaling (e.g. dataframe with groups columns).
                 Defaults to None.
             inverse (bool, optional): if to conduct inverse transformation. Defaults to False.
+            group_id (bool, optional): If the passed name refers to a group id (different encoders are used for these).
+                Defaults to False.
 
         Returns:
             np.ndarray: (de/en)coded/(de)scaled values
         """
+        if group_id:
+            name = self._group_ids_mapping[name]
         # remaining categories
-        if name in set(self.flat_categoricals + self.group_ids):
+        if name in set(self.flat_categoricals + self.group_ids + self._group_ids):
             name = self.variable_to_group_mapping.get(name, name)  # map name to encoder
             encoder = self.categorical_encoders[name]
             if encoder is None:
@@ -513,14 +610,15 @@ class TimeSeriesDataSet(Dataset):
                 transform = scaler.transform
             if isinstance(scaler, GroupNormalizer):
                 return transform(values, data)
-            elif isinstance(scaler, EncoderNormalizer):
+            elif isinstance(scaler, EncoderNormalizer) and name == self.target:
                 return transform(values)
             else:
                 if isinstance(values, pd.Series):
                     values = values.to_frame()
+                    return np.asarray(transform(values)).reshape(-1)
                 else:
                     values = values.reshape(-1, 1)
-                return transform(values).reshape(-1)
+                    return transform(values).reshape(-1)
 
         return values  # fallback
 
@@ -536,7 +634,7 @@ class TimeSeriesDataSet(Dataset):
                 time index
         """
 
-        index = torch.tensor(data[self.group_ids].to_numpy(np.long), dtype=torch.long)
+        index = torch.tensor(data[self._group_ids].to_numpy(np.long), dtype=torch.long)
         time = torch.tensor(data["__time_idx__"].to_numpy(np.long), dtype=torch.long)
 
         categorical = torch.tensor(data[self.flat_categoricals].to_numpy(np.long), dtype=torch.long)
@@ -612,7 +710,9 @@ class TimeSeriesDataSet(Dataset):
             Dict[str, Any]: dictionary of parameters
         """
         kwargs = {
-            name: getattr(self, name) for name in inspect.signature(self.__class__).parameters.keys() if name != "data"
+            name: getattr(self, name)
+            for name in inspect.signature(self.__class__.__init__).parameters.keys()
+            if name not in ["data", "self"]
         }
         kwargs["categorical_encoders"] = self.categorical_encoders
         kwargs["scalers"] = self.scalers
@@ -694,7 +794,7 @@ class TimeSeriesDataSet(Dataset):
         Returns:
             pd.DataFrame: index dataframe
         """
-        g = data.groupby(self.group_ids, observed=True)
+        g = data.groupby(self._group_ids, observed=True)
 
         df_index_first = g["__time_idx__"].transform("nth", 0).to_frame("time_first")
         df_index_last = g["__time_idx__"].transform("nth", -1).to_frame("time_last")
@@ -754,17 +854,20 @@ class TimeSeriesDataSet(Dataset):
             # choose longest sequence
             df_index = df_index.loc[df_index.groupby("group_id").sequence_length.idxmax()]
 
-        assert len(df_index) > 0, "filters should not remove entries"
         # check that all groups/series have at least one entry in the index
         if not group_ids.isin(df_index.group_id).all():
-            missing_groups = data.loc[~group_ids.isin(df_index.group_id), self.group_ids].drop_duplicates()
+            missing_groups = data.loc[~group_ids.isin(df_index.group_id), self._group_ids].drop_duplicates()
             # decode values
-            for name in missing_groups.columns:
-                missing_groups[name] = self.transform_values(name, missing_groups[name], inverse=True)
-            raise ValueError(
-                f"Min encoder length and/or min prediction length is too large for {len(missing_groups)} series/group. "
-                f"First 10 examples: {list(missing_groups.iloc[:10].to_dict(orient='index').values())}"
+            for name, id in self._group_ids_mapping.items():
+                missing_groups[id] = self.transform_values(name, missing_groups[id], inverse=True, group_id=True)
+            warnings.warn(
+                "Min encoder length and/or min_prediction_idx and/or min prediction length is too large for "
+                f"{len(missing_groups)} series/groups which therefore are not present in the dataset index. "
+                "This means no predictions can be made for those series"
+                f"First 10 removed groups: {list(missing_groups.iloc[:10].to_dict(orient='index').values())}",
+                UserWarning,
             )
+        assert len(df_index) > 0, "filters should not remove entries"
 
         return df_index
 
@@ -962,6 +1065,14 @@ class TimeSeriesDataSet(Dataset):
                 (encoder_length - 0.5 * self.max_encoder_length) / self.max_encoder_length * 2.0
             )
 
+        # rescale covariates
+        for name, scaler in self.scalers.items():
+            if name != self.target and name in self.reals and isinstance(scaler, EncoderNormalizer):
+                # fit and transform
+                pos = self.reals.index(name)
+                scaler.fit(data_cont[:encoder_length, pos])
+                data_cont[:, pos] = scaler.transform(data_cont[:, pos])
+
         # rescale target
         if self.target_normalizer is not None and isinstance(self.target_normalizer, EncoderNormalizer):
             # fit and transform
@@ -1057,7 +1168,10 @@ class TimeSeriesDataSet(Dataset):
 
         # target
         target = rnn.pad_sequence([batch[1] for batch in batches], batch_first=True)
-        target_scale = torch.tensor([batch[0]["target_scale"] for batch in batches], dtype=torch.float)
+        if isinstance(batches[0][0]["target_scale"], torch.Tensor):  # stack tensor
+            target_scale = torch.stack([batch[0]["target_scale"] for batch in batches])
+        else:  # convert to tensor
+            target_scale = torch.tensor([batch[0]["target_scale"] for batch in batches], dtype=torch.float)
 
         return (
             dict(
@@ -1082,6 +1196,8 @@ class TimeSeriesDataSet(Dataset):
         """
         Get dataloader from dataset.
 
+        The
+
         Args:
             train (bool, optional): if dataloader is used for training or prediction
                 Will shuffle and drop last batch if True. Defaults to True.
@@ -1089,7 +1205,10 @@ class TimeSeriesDataSet(Dataset):
             batch_sampler (Union[Sampler, str]): batch sampler or string. One of
 
                 * "synchronized": ensure that samples in decoder are aligned in time. Does not support missing
-                  values in dataset.
+                  values in dataset. This makes only sense if the underlying algorithm makes use of values aligned
+                  in time.
+                * PyTorch Sampler instance: any PyTorch sampler, e.g. the WeightedRandomSampler()
+                * None: samples are taken randomly from times series.
 
             **kwargs: additional arguments to ``DataLoader()``
 
@@ -1150,18 +1269,18 @@ class TimeSeriesDataSet(Dataset):
             **kwargs,
         )
 
-    def x_to_index(self, x) -> pd.DataFrame:
+    def x_to_index(self, x: Dict[str, torch.Tensor]) -> pd.DataFrame:
         """
         Decode dataframe index from x.
 
         Returns:
             dataframe with time index column for first prediction and group ids
         """
-        index_data = {self.time_idx: x["decoder_time_idx"][:, 0]}
+        index_data = {self.time_idx: x["decoder_time_idx"][:, 0].cpu()}
         for id in self.group_ids:
-            index_data[id] = x["groups"][:, self.group_ids.index(id)]
+            index_data[id] = x["groups"][:, self.group_ids.index(id)].cpu()
             # decode if possible
-            index_data[id] = self.transform_values(id, index_data[id], inverse=True)
+            index_data[id] = self.transform_values(id, index_data[id], inverse=True, group_id=True)
         index = pd.DataFrame(index_data)
         return index
 
