@@ -4,9 +4,12 @@ Implementation of metrics for (mulit-horizon) timeseries forecasting.
 from typing import Dict, List, Union
 import warnings
 
+import numpy as np
 from pytorch_lightning.metrics import Metric as LightningMetric
 import scipy.stats
+from sklearn.base import BaseEstimator
 import torch
+from torch import distributions
 import torch.nn.functional as F
 from torch.nn.utils import rnn
 
@@ -612,3 +615,258 @@ class MASE(MultiHorizonMetric):
         scaling = diffs.sum(1) / total_lengths + eps
 
         return scaling
+
+
+class DistributionLoss(MultiHorizonMetric):
+    """
+    DistributionLoss base class.
+
+    Class should be inherited for all distribution losses, i.e. if a network predicts
+    the parameters of a probability distribution, DistributionLoss can be used to
+    score those parameters and calculate loss for given true values.
+
+    Define two class attributes in a child class:
+
+    Attributes:
+        distribution_class (distributions.Distribution): torch probability distribution
+        distribution_arguments (List[str]): list of parameter names for the distribution
+
+    Further, implement the methods :py:meth:`~map_x_to_distribution` and :py:meth:`~rescale_parameters`.
+
+    """
+
+    distribution_class: distributions.Distribution
+    distribution_arguments: List[str]
+
+    def __init__(
+        self, name: str = None, quantiles: List[float] = [0.02, 0.1, 0.25, 0.5, 0.75, 0.9, 0.98], reduction="mean"
+    ):
+        """
+        Initialize metric
+
+        Args:
+            name (str): metric name. Defaults to class name.
+            quantiles (List[float], optional): quantiles for probability range. Defaults to None.
+            reduction (str, optional): Reduction, "none", "mean" or "sqrt-mean". Defaults to "mean".
+        """
+        super().__init__(name=name, quantiles=quantiles, reduction=reduction)
+
+    def map_x_to_distribution(self, x: torch.Tensor) -> distributions.Distribution:
+        """
+        Map the a tensor of parameters to a probability distribution.
+
+        Args:
+            x (torch.Tensor): parameters for probability distribution. Last dimension will index the parameters
+
+        Returns:
+            distributions.Distribution: torch probability distribution as defined in the
+                class attribute ``distribution_class``
+        """
+        raise NotImplementedError("implement this method")
+
+    def loss(self, y_pred: torch.Tensor, y_actual: torch.Tensor) -> torch.Tensor:
+        """
+        Calculate negative likelihood
+
+        Args:
+            y_pred: network output
+            y_actual: actual values
+
+        Returns:
+            torch.Tensor: metric value on which backpropagation can be applied
+        """
+        distribution = self.map_x_to_distribution(y_pred)
+        loss = -distribution.log_prob(y_actual)
+        return loss
+
+    def rescale_parameters(
+        self, parameters: torch.Tensor, target_scale: torch.Tensor, encoder: BaseEstimator
+    ) -> torch.Tensor:
+        """
+        Rescale normalized parameters into the scale required for the distribution.
+
+        Args:
+            parameters (torch.Tensor): normalized parameters (indexed by last dimension)
+            target_scale (torch.Tensor): scale of parameters (n_batch_samples x (center, scale))
+            encoder (BaseEstimator): original encoder that normalized the target in the first place
+
+        Returns:
+            torch.Tensor: parameters in real/not normalized space
+        """
+        return parameters
+
+    def to_prediction(self, y_pred: torch.Tensor) -> torch.Tensor:
+        """
+        Convert network prediction into a point prediction.
+
+        Args:
+            y_pred: prediction output of network (with ``output_type = samples``)
+
+        Returns:
+            torch.Tensor: mean prediction
+        """
+        return y_pred.mean(-1)
+
+    def sample_n(self, y_pred, n_samples: int) -> torch.Tensor:
+        """
+        Sample from distribution.
+
+        Args:
+            y_pred: prediction output of network
+            n_samples (int): number of samples to draw
+
+        Returns:
+            torch.Tensor: tensor where first dimensionare samples
+        """
+        dist = self.map_x_to_distribution(y_pred)
+        return dist.sample_n(n_samples)
+
+    def to_quantiles(self, y_pred: torch.Tensor) -> torch.Tensor:
+        """
+        Convert network prediction into a quantile prediction.
+
+        Args:
+            y_pred: prediction output of network (with ``output_type = samples``)
+
+        Returns:
+            torch.Tensor: prediction quantiles (last dimension)
+        """
+        samples = y_pred.size(-1)
+        quantiles = torch.stack(
+            [
+                torch.kthvalue(y_pred, int(samples * q), dim=-1)[0] if samples > 1 else y_pred[..., 0]
+                for q in self.quantiles
+            ],
+            dim=-1,
+        )
+        return quantiles
+
+
+class NormalDistributionLoss(DistributionLoss):
+    """
+    Normal distribution loss.
+
+    Requirements for original target normalizer:
+        * not normalized in log space (use :py:class:`~LogNormalDistributionLoss`)
+        * not coerced to be positive
+    """
+
+    distribution_class = distributions.Normal
+    distribution_arguments = ["loc", "scale"]
+
+    def map_x_to_distribution(self, x: torch.Tensor) -> distributions.Normal:
+        return self.distribution_class(loc=x[..., 0], scale=x[..., 1])
+
+    def rescale_parameters(
+        self, parameters: torch.Tensor, target_scale: torch.Tensor, encoder: BaseEstimator
+    ) -> torch.Tensor:
+        assert encoder.transformation not in ["log", "log1p"], "Use LogNormalDistributionLoss for log scaled data"
+        assert encoder.transformation not in [
+            "softplus",
+            "relu",
+        ], "Cannot use NormalDistributionLoss for positive data"
+        assert encoder.transformation not in ["logit"], "Cannot use bound transformation such as 'logit'"
+        loc = encoder(dict(prediction=parameters[..., 0], target_scale=target_scale))
+        scale = F.softplus(parameters[..., 1]) * target_scale[..., 1].unsqueeze(1)
+        return torch.stack([loc, scale], dim=-1)
+
+
+class NegativeBinomialDistributionLoss(DistributionLoss):
+    """
+    Negative binomial loss, e.g. for count data.
+
+    Requirements for original target normalizer:
+        * not centered normalization (only rescaled)
+    """
+
+    distribution_class = distributions.NegativeBinomial
+    distribution_arguments = ["mean", "shape"]
+
+    def map_x_to_distribution(self, x: torch.Tensor) -> distributions.NegativeBinomial:
+        mean = x[..., 0]
+        shape = x[..., 1]
+        r = 1.0 / shape
+        p = mean / (mean + r)
+        return self.distribution_class(total_count=r, probs=p)
+
+    def rescale_parameters(
+        self, parameters: torch.Tensor, target_scale: torch.Tensor, encoder: BaseEstimator
+    ) -> torch.Tensor:
+        assert not encoder.center, "NegativeBinomialDistributionLoss is not compatible with `center=True` normalization"
+        assert encoder.transformation not in ["logit"], "Cannot use bound transformation such as 'logit'"
+        if encoder.transformation in ["log", "log1p"]:
+            mean = torch.exp(parameters[..., 0] * target_scale[..., 1].unsqueeze(-1))
+            shape = (
+                F.softplus(torch.exp(parameters[..., 1]))
+                / torch.exp(target_scale[..., 1].unsqueeze(-1)).sqrt()  # todo: is this correct?
+            )
+        else:
+            mean = F.softplus(parameters[..., 0]) * target_scale[..., 1].unsqueeze(-1)
+            shape = F.softplus(parameters[..., 1]) / target_scale[..., 1].unsqueeze(-1).sqrt()
+        return torch.stack([mean, shape], dim=-1)
+
+
+class LogNormalDistributionLoss(DistributionLoss):
+    """
+    Log-normal loss.
+
+    Requirements for original target normalizer:
+        * normalized target in log space
+    """
+
+    distribution_class = distributions.LogNormal
+    distribution_arguments = ["loc", "scale"]
+
+    def map_x_to_distribution(self, x: torch.Tensor) -> distributions.LogNormal:
+        return self.distribution_class(loc=x[..., 0], scale=x[..., 1])
+
+    def rescale_parameters(
+        self, parameters: torch.Tensor, target_scale: torch.Tensor, encoder: BaseEstimator
+    ) -> torch.Tensor:
+        assert isinstance(encoder.transformation, str) and encoder.transformation in [
+            "log",
+            "log1p",
+        ], f"Log distribution requires log scaling but found `transformation={encoder.transform}`"
+
+        assert encoder.transformation not in ["logit"], "Cannot use bound transformation such as 'logit'"
+
+        scale = F.softplus(parameters[..., 1]) * target_scale[..., 1].unsqueeze(-1)
+        loc = parameters[..., 0] * target_scale[..., 1].unsqueeze(-1) + target_scale[..., 0].unsqueeze(-1)
+
+        return torch.stack([loc, scale], dim=-1)
+
+
+class BetaDistributionLoss(DistributionLoss):
+    """
+    Beta distribution loss for unit interval data.
+
+    Requirements for original target normalizer:
+        * logit transformation
+    """
+
+    distribution_class = distributions.Beta
+    distribution_arguments = ["mean", "shape"]
+
+    def map_x_to_distribution(self, x: torch.Tensor) -> distributions.Beta:
+        mean = x[..., 0]
+        shape = x[..., 1]
+        return self.distribution_class(concentration0=(1 - mean) * shape, concentration1=mean * shape)
+
+    def rescale_parameters(
+        self, parameters: torch.Tensor, target_scale: torch.Tensor, encoder: BaseEstimator
+    ) -> torch.Tensor:
+        assert encoder.transformation in ["logit"], "Beta distribution is only compatible with logit transformation"
+        assert encoder.center, "Beta distribution requires normalizer to center data"
+
+        scaled_mean = encoder(dict(prediction=parameters[..., 0], target_scale=target_scale))
+        # need to first transform target scale standard deviation in logit space to real space
+        # we assume a normal distribution in logit space (we used a logit transform and a standard scaler)
+        # and know that the variance of the beta distribution is limited by `scaled_mean * (1 - scaled_mean)`
+        mean_derivative = scaled_mean * (1 - scaled_mean)
+
+        # we can approximate variance as
+        # torch.pow(torch.tanh(target_scale[..., 1].unsqueeze(1) * torch.sqrt(mean_derivative)), 2) * mean_derivative
+        # shape is (positive) parameter * mean_derivative / var
+        shape_scaler = torch.pow(torch.tanh(target_scale[..., 1].unsqueeze(1) * torch.sqrt(mean_derivative)), 2)
+        scaled_shape = F.softplus(parameters[..., 1]) / shape_scaler
+        return torch.stack([scaled_mean, scaled_shape], dim=-1)
