@@ -11,6 +11,7 @@ import numpy as np
 import pandas as pd
 from pytorch_lightning import LightningModule
 from pytorch_lightning.utilities.parsing import get_init_args
+import scipy.stats
 import torch
 import torch.nn as nn
 from torch.nn.utils import rnn
@@ -22,7 +23,7 @@ from pytorch_forecasting.data import TimeSeriesDataSet
 from pytorch_forecasting.data.encoders import EncoderNormalizer, GroupNormalizer
 from pytorch_forecasting.metrics import MASE, SMAPE, Metric
 from pytorch_forecasting.optim import Ranger
-from pytorch_forecasting.utils import groupby_apply
+from pytorch_forecasting.utils import get_embedding_size, groupby_apply
 
 
 class BaseModel(LightningModule):
@@ -33,6 +34,29 @@ class BaseModel(LightningModule):
     The ``forward()`` method should return a dictionary with at least the entry ``prediction`` and
     ``target_scale`` that contains the network's output.
 
+    The idea of the base model is that common methods do not have to be re-implemented for every new architecture.
+    The class is a [LightningModule](https://pytorch-lightning.readthedocs.io/en/latest/lightning_module.html)
+    and follows its conventions. However, there are important additions:
+
+        * You need to specify a ``loss`` attribute that stores the function to calculate the
+          :py:class:`~pytorch_forecasting.metrics.MultiHorizonLoss` for backpropagation.
+        * The :py:meth:`~BaseModel.from_dataset` method can be used to initialize a network using the specifications
+          of a dataset. Often, parameters such as the number of features can be easily deduced from the dataset.
+          Further, the method will also store how to rescale normalized predictions into the unnormalized prediction
+          space. Override it to pass additional arguments to the __init__ method of your network that depend on your
+          dataset.
+        * The :py:meth:`~BaseModel.transform_output` method rescales the network output using the target normalizer
+          from thedataset.
+        * The :py:meth:`~BaseModel.step` method takes care of calculating the loss, logging additional metrics defined
+          in the ``logging_metrics`` attribute and plots of sample predictions. You can override this method to add
+          custom interpretations or pass extra arguments to the networks forward method.
+        * The :py:meth:`~BaseModel.epoch_end` method can be used to calculate summaries of each epoch such as
+          statistics on the encoder length, etc.
+        * The :py:meth:`~BaseModel.predict` method makes predictions using a dataloader or dataset. Override it if you
+          need to pass additional arguments to ``forward`` by default.
+
+    To implement your own architecture, it is best to look at existing ones to understand what might be a good approach.
+
     Example:
 
         .. code-block:: python
@@ -59,6 +83,7 @@ class BaseModel(LightningModule):
         loss: Metric = SMAPE(),
         logging_metrics: nn.ModuleList = nn.ModuleList([]),
         reduce_on_plateau_patience: int = 1000,
+        reduce_on_plateau_min_lr: float = 1e-5,
         weight_decay: float = 0.0,
         monotone_constaints: Dict[str, int] = {},
         output_transformer: Callable = None,
@@ -80,6 +105,8 @@ class BaseModel(LightningModule):
                 Defaults to [].
             reduce_on_plateau_patience (int): patience after which learning rate is reduced by a factor of 10. Defaults
                 to 1000
+            reduce_on_plateau_min_lr (float): minimum learning rate for reduce on plateua learning rate scheduler.
+                Defaults to 1e-5
             weight_decay (float): weight decay. Defaults to 0.0.
             monotone_constaints (Dict[str, int]): dictionary of monotonicity constraints for continuous decoder
                 variables mapping
@@ -88,7 +115,7 @@ class BaseModel(LightningModule):
                 This constraint significantly slows down training. Defaults to {}.
             output_transformer (Callable): transformer that takes network output and transforms it to prediction space.
                 Defaults to None which is equivalent to ``lambda out: out["prediction"]``.
-            optimizer (str): optimizer
+            optimizer (str): Optimizer, "ranger", "adam" or "adamw". Defaults to "ranger".
         """
         super().__init__()
         # update hparams
@@ -144,13 +171,13 @@ class BaseModel(LightningModule):
     def validation_epoch_end(self, outputs):
         self.epoch_end(outputs, label="val")
 
-    def step(self, x: Dict[str, torch.Tensor], y: torch.Tensor, batch_idx: int, label="train"):
+    def step(self, x: Dict[str, torch.Tensor], y: torch.Tensor, batch_idx: int, label="train", **kwargs):
         """
         Run for each train/val step.
         """
         # pack y sequence if different encoder lengths exist
         if (x["decoder_lengths"] < x["decoder_lengths"].max()).any():
-            y = rnn.pack_padded_sequence(y, lengths=x["decoder_lengths"], batch_first=True, enforce_sorted=False)
+            y = rnn.pack_padded_sequence(y, lengths=x["decoder_lengths"].cpu(), batch_first=True, enforce_sorted=False)
 
         if label == "train" and len(self.hparams.monotone_constaints) > 0:
             # calculate gradient with respect to continous decoder features
@@ -159,7 +186,7 @@ class BaseModel(LightningModule):
                 "To use monotone constraints, wrap model and training in context "
                 "`torch.backends.cudnn.flags(enable=False)`"
             )
-            out = self(x)
+            out = self(x, **kwargs)
             out["prediction"] = self.transform_output(out)
             prediction = out["prediction"]
 
@@ -193,7 +220,7 @@ class BaseModel(LightningModule):
 
             loss = loss * (1 + monotinicity_loss)
         else:
-            out = self(x)
+            out = self(x, **kwargs)
             out["prediction"] = self.transform_output(out)
 
             # calculate loss
@@ -205,8 +232,30 @@ class BaseModel(LightningModule):
             else:
                 loss = self.loss(prediction, y)
 
+        # log
+        self._log_metrics(x, y, out, label=label)
+        if self.log_interval(label == "train") > 0:
+            self._log_prediction(x, out, batch_idx, label=label)
+        log = {"loss": loss, "n_samples": x["decoder_lengths"].size(0)}
+
+        return log, out
+
+    def _log_metrics(
+        self, x: Dict[str, torch.Tensor], y: torch.Tensor, out: Dict[str, torch.Tensor], label: str = "train"
+    ) -> None:
+        """
+        Log metrics every training/validation step
+
+        [extended_summary]
+
+        Args:
+            x (Dict[str, torch.Tensor]): [description]
+            y (torch.Tensor): [description]
+            out (Dict[str, torch.Tensor]): [description]
+            label (str, optional): [description]. Defaults to "train".
+        """
         # logging losses
-        y_hat_detached = prediction.detach()
+        y_hat_detached = out["prediction"].detach()
         y_hat_point_detached = self.loss.to_prediction(y_hat_detached)
         for metric in self.logging_metrics:
             if isinstance(metric, MASE):
@@ -216,10 +265,6 @@ class BaseModel(LightningModule):
             else:
                 loss_value = metric(y_hat_point_detached, y)
             self.log(f"{label}_{metric.name}", loss_value, on_step=label == "train", on_epoch=True)
-        log = {"loss": loss, "n_samples": x["decoder_lengths"].size(0)}
-        if self.log_interval(label == "train") > 0:
-            self._log_prediction(x, out, batch_idx, label=label)
-        return log, out
 
     def forward(self, x: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
         """
@@ -248,7 +293,7 @@ class BaseModel(LightningModule):
         else:
             return self.hparams.log_val_interval
 
-    def _log_prediction(self, x, out, batch_idx, label="train"):
+    def _log_prediction(self, x, out, batch_idx, label="train") -> None:
         # log single prediction figure
         log_interval = self.log_interval(label == "train")
         if (batch_idx % log_interval == 0 or log_interval < 1.0) and log_interval > 0:
@@ -276,7 +321,7 @@ class BaseModel(LightningModule):
         x: Dict[str, torch.Tensor],
         out: Dict[str, torch.Tensor],
         idx: int = 0,
-        add_loss_to_title: Union[Metric, bool] = False,
+        add_loss_to_title: Union[Metric, torch.Tensor, bool] = False,
         show_future_observed: bool = True,
         ax=None,
     ) -> plt.Figure:
@@ -287,7 +332,8 @@ class BaseModel(LightningModule):
             x: network input
             out: network output
             idx: index of prediction to plot
-            add_loss_to_title: if to add loss to title or loss function to calculate. Default to False.
+            add_loss_to_title: if to add loss to title or loss function to calculate. Can be either metrics,
+                bool indicating if to use loss metric or tensor which contains losses for all samples. Default to False.
             show_future_observed: if to show actuals for future. Defaults to True.
             ax: matplotlib axes to plot on
 
@@ -356,16 +402,22 @@ class BaseModel(LightningModule):
                     c=pred_color,
                     capsize=1.0,
                 )
-        if add_loss_to_title:
+        if add_loss_to_title is not False:
             if isinstance(add_loss_to_title, bool):
                 loss = self.loss
-            else:
+            elif isinstance(add_loss_to_title, torch.Tensor):
+                loss = add_loss_to_title.detach()[idx].item()
+            elif isinstance(add_loss_to_title, Metric):
                 loss = add_loss_to_title
                 loss.quantiles = self.loss.quantiles
+            else:
+                raise ValueError(f"add_loss_to_title '{add_loss_to_title}'' is unkown")
             if isinstance(loss, MASE):
                 loss_value = loss(y_hat[None], y[-n_pred:][None], y[:n_pred][None])
-            else:
+            elif isinstance(loss, Metric):
                 loss_value = loss(y_hat[None], y[-n_pred:][None])
+            else:
+                loss_value = loss
             ax.set_title(f"Loss {loss_value:.3g}")
         ax.set_xlabel("Time index")
         fig.legend()
@@ -451,6 +503,7 @@ class BaseModel(LightningModule):
                         factor=0.1,
                         patience=self.hparams.reduce_on_plateau_patience,
                         cooldown=self.hparams.reduce_on_plateau_patience,
+                        min_lr=self.hparams.reduce_on_plateau_min_lr,
                     ),
                     "monitor": "val_loss",  # Default: val_loss
                     "interval": "epoch",
@@ -510,6 +563,7 @@ class BaseModel(LightningModule):
         fast_dev_run: bool = False,
         show_progress_bar: bool = False,
         return_x: bool = False,
+        **kwargs,
     ):
         """
         predict dataloader
@@ -525,6 +579,7 @@ class BaseModel(LightningModule):
             fast_dev_run: if to only return results of first batch
             show_progress_bar: if to show progress bar. Defaults to False.
             return_x: if to return network inputs
+            **kwargs: additional arguments to network's forward method
 
         Returns:
             output, x, index, decoder_lengths: some elements might not be present depending on what is configured
@@ -555,10 +610,10 @@ class BaseModel(LightningModule):
                 # move data to appropriate device
                 for name in x.keys():
                     if x[name].device != self.device:
-                        x[name].to(self.device)
+                        x[name] = x[name].to(self.device)
 
                 # make prediction
-                out = self(x)  # raw output is dictionary
+                out = self(x, **kwargs)  # raw output is dictionary
                 out["prediction"] = self.transform_output(out)
 
                 lengths = x["decoder_lengths"]
@@ -600,7 +655,14 @@ class BaseModel(LightningModule):
         elif mode == "raw":
             output_cat = {}
             for name in output[0].keys():
-                output_cat[name] = torch.cat([out[name] for out in output], dim=0)
+                v0 = output[0][name]
+                if isinstance(v0, torch.Tensor):
+                    output_cat[name] = torch.cat([out[name] for out in output], dim=0)
+                else:
+                    try:
+                        output_cat[name] = np.concatenate([out[name] for out in output], axis=0)
+                    except ValueError:
+                        output_cat[name] = [out[name] for out in output]
             output = output_cat
 
         # generate output
@@ -713,9 +775,9 @@ class BaseModel(LightningModule):
         return results
 
 
-class CovariatesMixin:
+class BaseModelWithCovariates(BaseModel):
     """
-    Model mix-in for additional methods using covariates.
+    Model with additional methods using covariates.
 
     Assumes the following hyperparameters:
 
@@ -725,8 +787,46 @@ class CovariatesMixin:
         embedding_sizes: dictionary mapping (string) indices to tuple of number of categorical classes and
             embedding size
         embedding_labels: dictionary mapping (string) indices to list of categorical labels
+        static_categoricals: integer of positions of static categorical variables
+        static_reals: integer of positions of static continuous variables
+        time_varying_categoricals_encoder: integer of positions of categorical variables for encoder
+        time_varying_categoricals_decoder: integer of positions of categorical variables for decoder
+        time_varying_reals_encoder: integer of positions of continuous variables for encoder
+        time_varying_reals_decoder: integer of positions of continuous variables for decoder
     """
 
+    @property
+    def reals(self) -> List[str]:
+        return list(
+            set(
+                self.hparams.static_reals
+                + self.hparams.time_varying_reals_encoder
+                + self.hparams.time_varying_reals_decoder
+            )
+        )
+
+    @property
+    def categoricals(self) -> List[str]:
+        return list(
+            set(
+                self.hparams.static_categoricals
+                + self.hparams.time_varying_categoricals_encoder
+                + self.hparams.time_varying_categoricals_decoder
+            )
+        )
+
+    @property
+    def static_variables(self) -> List[str]:
+        return self.hparams.static_categoricals + self.hparams.static_reals
+
+    @property
+    def encoder_variables(self) -> List[str]:
+        return self.hparams.time_varying_categoricals_encoder + self.hparams.time_varying_reals_encoder
+
+    @property
+    def decoder_variables(self) -> List[str]:
+        return self.hparams.time_varying_categoricals_decoder + self.hparams.time_varying_reals_decoder
+
     @property
     def categorical_groups_mapping(self) -> Dict[str, str]:
         groups = {}
@@ -734,6 +834,68 @@ class CovariatesMixin:
             groups.update({name: group_name for name in sublist})
         return groups
 
+    @classmethod
+    def from_dataset(
+        cls,
+        dataset: TimeSeriesDataSet,
+        allowed_encoder_known_variable_names: List[str] = None,
+        **kwargs,
+    ) -> LightningModule:
+        """
+        Create model from dataset and set parameters related to covariates.
+
+        Args:
+            dataset: timeseries dataset
+            allowed_encoder_known_variable_names: List of known variables that are allowed in encoder, defaults to all
+            **kwargs: additional arguments such as hyperparameters for model (see ``__init__()``)
+
+        Returns:
+            LightningModule
+        """
+        # assert fixed encoder and decoder length for the moment
+        if allowed_encoder_known_variable_names is None:
+            allowed_encoder_known_variable_names = (
+                dataset.time_varying_known_categoricals + dataset.time_varying_known_reals
+            )
+
+        # embeddings
+        embedding_labels = {
+            name: encoder.classes_
+            for name, encoder in dataset.categorical_encoders.items()
+            if name in dataset.categoricals
+        }
+        embedding_paddings = dataset.dropout_categoricals
+        # determine embedding sizes based on heuristic
+        embedding_sizes = {
+            name: (len(encoder.classes_), get_embedding_size(len(encoder.classes_)))
+            for name, encoder in dataset.categorical_encoders.items()
+            if name in dataset.categoricals
+        }
+        embedding_sizes.update(kwargs.get("embedding_sizes", {}))
+        kwargs.setdefault("embedding_sizes", embedding_sizes)
+
+        new_kwargs = dict(
+            static_categoricals=dataset.static_categoricals,
+            time_varying_categoricals_encoder=[
+                name for name in dataset.time_varying_known_categoricals if name in allowed_encoder_known_variable_names
+            ]
+            + dataset.time_varying_unknown_categoricals,
+            time_varying_categoricals_decoder=dataset.time_varying_known_categoricals,
+            static_reals=dataset.static_reals,
+            time_varying_reals_encoder=[
+                name for name in dataset.time_varying_known_reals if name in allowed_encoder_known_variable_names
+            ]
+            + dataset.time_varying_unknown_reals,
+            time_varying_reals_decoder=dataset.time_varying_known_reals,
+            x_reals=dataset.reals,
+            x_categoricals=dataset.flat_categoricals,
+            embedding_labels=embedding_labels,
+            embedding_paddings=embedding_paddings,
+            categorical_groups=dataset.variable_groups,
+        )
+        new_kwargs.update(kwargs)
+        return super().from_dataset(dataset, **new_kwargs)
+
     def calculate_prediction_actual_by_variable(
         self,
         x: Dict[str, torch.Tensor],
@@ -741,16 +903,19 @@ class CovariatesMixin:
         normalize: bool = True,
         bins: int = 95,
         std: float = 2.0,
+        log_scale: bool = None,
     ) -> Dict[str, Dict[str, torch.Tensor]]:
         """
         Calculate predictions and actuals by variable averaged by ``bins`` bins spanning from ``-std`` to ``+std``
 
         Args:
             x: input as ``forward()``
-            y_pred: predictions obtained by ``self.transform_output(self(x))``
+            y_pred: predictions obtained by ``self.transform_output(self(x, **kwargs))``
             normalize: if to return normalized averages, i.e. mean or sum of ``y``
             bins: number of bins to calculate
             std: number of standard deviations for standard scaled continuous variables
+            log_scale (str, optional): if to plot in log space. If None, determined based on skew of values.
+                Defaults to None.
 
         Returns:
             dictionary that can be used to plot averages with :py:meth:`~plot_prediction_actual_by_variable`
@@ -766,10 +931,13 @@ class CovariatesMixin:
         # select valid y values
         y_flat = x["decoder_target"][mask]
         y_pred_flat = y_pred[mask]
-        log_y = self.dataset_parameters["target_normalizer"] is not None and getattr(
-            self.dataset_parameters["target_normalizer"], "log_scale", False
-        )
-        if log_y:
+
+        # determine in which average in log-space to transform data
+        if log_scale is None:
+            skew = torch.mean(((y_flat - torch.mean(y_flat)) / torch.std(y_flat)) ** 3)
+            log_scale = skew > 1.6
+
+        if log_scale:
             y_flat = torch.log(y_flat + 1e-8)
             y_pred_flat = torch.log(y_pred_flat + 1e-8)
 
@@ -836,7 +1004,7 @@ class CovariatesMixin:
                 averages_actual[name] /= support[name].clamp(min=1)
                 averages_prediction[name] /= support[name].clamp(min=1)
 
-        if log_y:  # reverse log scaling
+        if log_scale:
             for name in support.keys():
                 averages_actual[name] = torch.exp(averages_actual[name])
                 averages_prediction[name] = torch.exp(averages_prediction[name])
@@ -848,7 +1016,7 @@ class CovariatesMixin:
         }
 
     def plot_prediction_actual_by_variable(
-        self, data: Dict[str, Dict[str, torch.Tensor]], name: str = None, ax=None
+        self, data: Dict[str, Dict[str, torch.Tensor]], name: str = None, ax=None, log_scale: bool = None
     ) -> Union[Dict[str, plt.Figure], plt.Figure]:
         """
         Plot predicions and actual averages by variables
@@ -858,6 +1026,8 @@ class CovariatesMixin:
                 :py:meth:`~calculate_prediction_actual_by_variable`
             name (str, optional): name of variable for which to plot actuals vs predictions. Defaults to None which
                 means returning a dictionary of plots for all variables.
+            log_scale (str, optional): if to plot in log space. If None, determined based on skew of values.
+                Defaults to None.
 
         Raises:
             ValueError: if the variable name is unkown
@@ -891,17 +1061,19 @@ class CovariatesMixin:
             bins = values_actual.size
             support = data["support"][name].cpu().numpy()
 
-            if self.dataset_parameters["target_normalizer"] is not None and getattr(
-                self.dataset_parameters["target_normalizer"], "log_scale", False
-            ):
-                ax.set_yscale("log")
-
             # only display values where samples were observed
             support_non_zero = support > 0
             support = support[support_non_zero]
             values_actual = values_actual[support_non_zero]
             values_prediction = values_prediction[support_non_zero]
 
+            # determine if to display results in log space
+            if log_scale is None:
+                log_scale = scipy.stats.skew(values_actual) > 1.6
+
+            if log_scale:
+                ax.set_yscale("log")
+
             # plot averages
             if name in self.hparams.x_reals:
                 # create x
@@ -957,4 +1129,25 @@ class CovariatesMixin:
 
 
 class AutoRegressiveBaseModel(BaseModel):
+    @classmethod
+    def from_dataset(
+        cls,
+        dataset: TimeSeriesDataSet,
+        **kwargs,
+    ) -> LightningModule:
+        """
+        Create model from dataset.
+
+        Args:
+            dataset: timeseries dataset
+            **kwargs: additional arguments such as hyperparameters for model (see ``__init__()``)
+
+        Returns:
+            LightningModule
+        """
+        kwargs.setdefault("target", dataset.target)
+        return super().from_dataset(dataset, **kwargs)
+
+
+class AutoRegressiveBaseModelWithCovariates(BaseModelWithCovariates, AutoRegressiveBaseModel):
     pass
